# -*- coding: utf-8 -*-
"""
Created on Wed Apr 29 08:21:46 2015.

@author: mel
"""

from anypytools.macroutils import (MacroCommand, Load, SetValue, SetValue_random,
                                   Dump, SaveDesign, LoadDesign, SaveValues,
                                   LoadValues, UpdateValues, OperationRun, SaveData)

__all__ = [
    'MacroCommand',
    'Load',
    'SetValue',
    'SetValue_random',
    'Dump',
    'SaveDesign',
    'LoadDesign',
    'SaveValues',
    'LoadValues',
    'UpdateValues',
    'OperationRun',
    'SaveData'
]



# -*- coding: utf-8 -*-
"""
Created on Sun Sep  7 13:25:38 2014.

@author: Morten
"""
# Python 2/3 compatibility imports
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import (ascii, bytes, chr, dict, filter, hex, input,  # noqa
                      int, map, next, oct, open, pow, range, round,
                      str, super, zip)

from past.builtins import basestring as string_types  # noqa


# Standard lib imports
import os
import re
import sys
import xml
import copy
import errno
import shutil
import pprint as _pprint
import logging
import textwrap
import datetime
import warnings
import platform
import subprocess
import collections
from ast import literal_eval
from _thread import get_ident as _get_ident

# external imports
import numpy as np
from future.utils import raise_from


logger = logging.getLogger('abt.anypytools')


# This hacks pprint to always return strings witout u' prefix
# important when running doctest in both python 2 og python 3
class Py3kPrettyPrinter(_pprint.PrettyPrinter):
    """Pretty printer which creates strings without the u' prefix."""

    def format(self, object, context, maxlevels, level):
        try:
            if isinstance(object, unicode):
                rep = u"'" + object + u"'"
                return (rep.encode('utf8'), True, False)
        except NameError:
            pass
        return _pprint.PrettyPrinter.format(self, object, context, maxlevels, level)


def py3k_pprint(s):
    printer = Py3kPrettyPrinter(width=110)
    printer.pprint(s)


pprint = py3k_pprint


def run_from_ipython():
    try:
        __IPYTHON__
        return True
    except NameError:
        return False


ANYBODYCON_VERSION_RE = re.compile(r'.*version\s:\s(?P<version>(?P<v1>\d)\.\s(?P<v2>\d)'
                                   r'\.\s(?P<v3>\d)\.\s(?P<build>\d+)\s\((?P<arc>.*)\))')

ON_WINDOWS = (platform.system() == 'Windows')


def anybodycon_version(anybodyconpath):
    """Return the AnyBodyCon version."""
    anybodyconpath = anybodyconpath or get_anybodycon_path()
    if anybodyconpath is None:
        return '0.0.0'
    try:
        out = subprocess.check_output(
            [anybodyconpath, '-ni'], universal_newlines=True)
    except (subprocess.CalledProcessError, FileNotFoundError):
        return None
    m = ANYBODYCON_VERSION_RE.search(out)
    if m is not None:
        return m.groupdict()['version']


AMMR_VERSION_RE = re.compile(r'.*AMMR_VERSION\s"(?P<version>.*)"')


def ammr_any_version(fpath):
    with open(fpath) as f:
        out = f.read()
    match = AMMR_VERSION_RE.search(out)
    if match:
        return match.groupdict()['version']
    else:
        return "Unknown AMMR version"


def wraptext(elem, initial_indent='', subsequent_indent=None):
    """Wraps text to fit the terminal window."""
    width = 120
    if sys.version_info.major == 3:
        width = max(width, shutil.get_terminal_size().columns - 1)
    subsequent_indent = subsequent_indent or initial_indent
    return textwrap.fill(elem, width,
                         initial_indent=initial_indent,
                         subsequent_indent=subsequent_indent)


def amm_xml_version(fpath):
    try:
        tree = xml.etree.ElementTree.parse(fpath)
        version = tree.getroot()
        v1, v2, v3 = version.find('v1').text, version.find(
            'v2').text, version.find('v3').text
        return "{}.{}.{}".format(v1, v2, v3)
    except Exception:
        vstring = "Unknown AMMR version"
    return vstring


def find_ammr_path(folder=None):
    """Return the root AMMR path if possible .

    The function will walk up a directory tree looking
    for a ammr_verion.any file to parse.
    """
    folder = folder or os.getcwd()
    version_files = ('AMMR.version.any', 'AMMR.version.xml')
    for basedir, dirs, files in walk_up(folder):
        if any(fn in files for fn in version_files):
            return basedir
    else:
        return None


def get_ammr_version(folder=None):
    """Return the AMMR version if possible.

    The function will walk up a directory tree looking
    for a ammr_verion.any file to parse.
    """
    folder = folder or os.getcwd()
    any_version_file = 'AMMR.version.any'
    xml_version_file = 'AMMR.version.xml'
    files = os.listdir(folder)
    if any_version_file in files:
        return ammr_any_version(os.path.join(folder, any_version_file))
    elif xml_version_file in files:
        return amm_xml_version(os.path.join(folder, xml_version_file))
    else:
        return ""


def walk_up(bottom):
    """Mimic os.walk, but walk 'up' instead of down the directory tree."""
    bottom = os.path.realpath(bottom)
    # get files in current dir
    names = os.listdir(bottom)
    dirs, nondirs = [], []
    for name in names:
        if os.path.isdir(os.path.join(bottom, name)):
            dirs.append(name)
        else:
            nondirs.append(name)
    yield bottom, dirs, nondirs
    new_path = os.path.realpath(os.path.join(bottom, '..'))
    # see if we are at the top
    if new_path == bottom:
        return
    for x in walk_up(new_path):
        yield x


def get_current_time():
    return datetime.datetime.utcnow().strftime("%Y%m%d_%H%M%S")


def get_tag(project_name=None):
    info = get_git_commit_info(project_name)
    parts = [info['id'], get_current_time()]
    if info['dirty']:
        parts.append("uncommited-changes")
    return "_".join(parts)


def get_git_project_name():
    cmd = 'git config --local remote.origin.url'.split()
    try:
        output = subprocess.check_output(cmd, universal_newlines=True)
    except (subprocess.CalledProcessError, IOError):
        name = None
    else:
        name = re.findall(r'/([^/]*)\.git', output)[0]
    return name if name is not None else os.path.basename(os.getcwd())


def get_git_branch_info():
    """Return the branch name for git repository."""
    cmd = 'git rev-parse --abbrev-ref HEAD'
    try:
        branch = subprocess.check_output(cmd, universal_newlines=True).strip()
    except (subprocess.CalledProcessError, IOError) as e:
        branch = 'unknown'
    else:
        if branch == 'HEAD':
            return '(detached head)'
    return branch


def get_git_commit_info(project_name=None):
    dirty = False
    commit = 'unversioned'
    project_name = project_name or get_git_project_name()
    branch = get_git_branch_info()
    cmd = 'git describe --dirty --always --long --abbrev=6'.split()
    try:
        output = subprocess.check_output(cmd, universal_newlines=True).strip()
    except (subprocess.CalledProcessError, IOError) as e:
        pass
    else:
        try:
            output = output.split('-')
            if output[-1].strip() == 'dirty':
                dirty = True
                output.pop()
            commit = output[-1].strip('g')
        except Exception as e:
            commit = 'unknown'
    return dict(id=commit, dirty=dirty, project=project_name, branch=branch)


def _get_first_key_match(key, names):
    """Find the first partial match key match.

    If No match if found then key is returned unmodified.
    """
    if key in names:
        return key
    matching = [v for v in names if key in v]
    if not matching:
        # No match return original key.
        return key

    if len(matching) > 1:
        print('WARNING: "{}" key is not unique.'
              ' Using the first match'.format(key), file=sys.stderr)
        print('-> ' + matching[0], file=sys.stderr)
        for match in matching[1:]:
            print(' * ' + match, file=sys.stderr)

    return matching[0]


class AnyPyProcessOutputList(collections.MutableSequence):
    """List like class to wrap the output of model simulations.

    The class behaves as a normal list but provide
    extra function to easily access data.
    """

    def __init__(self, *args):
        self.list = list()
        for elem in args:
            self.extend(list(elem))

    def check(self, v):
        if not isinstance(v, collections.MutableSequence):
            v = [v]
        for e in v:
            if not isinstance(e, collections.OrderedDict):
                raise(TypeError(e))

    def __len__(self):
        return len(self.list)

    def __getitem__(self, i):
        if isinstance(i, string_types):
            # Find the entries where i matches the keys
            key = _get_first_key_match(i, self.list[0])
            try:
                data = np.array(
                    [super(AnyPyProcessOutput, e).__getitem__(key)
                     for e in self.list]
                )
            except KeyError as e:
                msg = " The key: '{}' is not present in all elements of the output."
                raise_from(KeyError(msg.format(key)), None)
            if data.dtype == np.dtype('O'):
                # Data will be stacked as an array of objects, if the length of the
                # time dimension is not consistant across simulations. Warn that some numpy
                # featurs will not be avaiable.
                warnings.warn('\n\The length of the time variable varies across macros. '
                              'Numpy does not support ragged arrays. Data is returned  '
                              'as an array of array objects')
            return data
        else:
            return type(self)(self.list[i]) if isinstance(i, slice) else self.list[i]

    def __delitem__(self, i):
        del self.list[i]

    def __setitem__(self, i, v):
        self.check(v)
        if isinstance(i, slice):
            self.list[i] = v
        else:
            self.list[i] = v

    def insert(self, i, v):
        self.check(v)
        self.list.insert(i, v)

    def __str__(self):
        return str(self.list)

    def __repr__(self):
        def create_repr(maxlength=500):
            repr_list = []
            for elem in self.list:
                if not isinstance(elem, AnyPyProcessOutput):
                    repr_list.append('  ' + _pprint.pformat(elem))
                    continue
                for line in elem._repr_gen(prefix=' '):
                    repr_list.append(line)
                    if maxlength and len(repr_list) > maxlength:
                        repr_list.append('  ...')
                        return repr_list
                if repr_list and not repr_list[-1].endswith(','):
                    repr_list[-1] = repr_list[-1] + ','

            if len(repr_list):
                repr_list[-1] = repr_list[-1].rstrip(',')
                repr_list[0] = '[' + repr_list[0][1:]
                repr_list[-1] = repr_list[-1] + ']'
            else:
                repr_list.append('[]')
            return repr_list

        repr_str = '\n'.join(create_repr(500))
        if repr_str.endswith('...'):
            np.set_printoptions(threshold=30)
            repr_str = '\n'.join(create_repr(1000))
            np.set_printoptions()
        return repr_str

    def filter(self, function):
        """Filter elements for whichfunction returns true."""
        return AnyPyProcessOutputList(filter(function, self))

    def to_dynd(self, **kwargs):
        try:
            from anypytools.blaze_converter import convert
            return convert(self.list, **kwargs)
        except ImportError as e:
            msg = ('The packages libdynd, dynd-python, datashape, '
                   'odo/into must be installed to convert data')
            raise raise_from(ImportError(msg), None)

    def shelve(self, filename, key='results'):
        import shelve
        db = shelve.open(filename)
        db[key] = self
        db.close()

    @classmethod
    def from_shelve(cls, filename, key='results'):
        import shelve
        db = shelve.open(filename)
        out = db[key]
        db.close()
        return out

    def tolist(self):
        """Return as native python types (list of dicts)."""
        return [
            {
                k: v.tolist() if isinstance(v, np.ndarray) else v
                for k, v in elem.items()
            }
            for elem in self
        ]


def _expand_short_path_name(short_path_name):
    from ctypes import create_unicode_buffer, windll
    BUFFER_SIZE = 500
    buffer = create_unicode_buffer(BUFFER_SIZE)
    get_long_path_name = windll.kernel32.GetLongPathNameW
    get_long_path_name(short_path_name, buffer, BUFFER_SIZE)
    long_path_name = buffer.value
    return long_path_name


def get_anybodycon_path():
    """Return the path to default AnyBody console application."""
    if not ON_WINDOWS:
        return None
    try:
        import winreg
    except ImportError:
        import _winreg as winreg
    try:
        abpath = winreg.QueryValue(winreg.HKEY_CLASSES_ROOT,
                                   'AnyBody.AnyScript\shell\open\command')
    except WindowsError:
        raise WindowsError('Could not locate AnyBody in registry')
    abpath = abpath.rsplit(' ', 1)[0].strip('"')
    abpath = os.path.join(os.path.dirname(abpath), 'AnyBodyCon.exe')
    if '~' in abpath:
        abpath = _expand_short_path_name(abpath)
    return abpath


def define2str(key, value=None):
    if isinstance(value, string_types):
        if value.startswith('"') and value.endswith('"'):
            defstr = '-def %s=---"\\"%s\\""' % (key,
                                                value[1:-1].replace('\\', '\\\\'))
        else:
            defstr = '-def %s="%s"' % (key, value)
    elif value is None:
        defstr = '-def %s=""' % (key)
    elif isinstance(value, float):
        defstr = '-def %s="%g"' % (key, value)
    else:
        defstr = '-def %s="%d"' % (key, value)
    return defstr


def path2str(key, path='.'):
    return '-p %s=---"%s"' % (key, path.replace('\\', '\\\\'))


def getsubdirs(toppath, search_string="."):
    """Find all directories below a given top path.

    Args:
        toppath: top directory when searching for sub directories
        search_string: Limit to directories matching the this regular expression
    Returns:
        List of directories
    """
    if not search_string:
        return [toppath]
    reg_prog = re.compile(search_string)
    dirlist = []
    if search_string == ".":
        dirlist.append(toppath)
    for root, dirs, files in os.walk(toppath):
        for fname in files:
            if reg_prog.search(os.path.join(root, fname)):
                dirlist.append(root)
                continue
    uniqueList = []
    for value in dirlist:
        if value not in uniqueList:
            uniqueList.append(value)
    return uniqueList


def array2anyscript(arr):
    """Format a numpy array as an anyscript variable."""
    def tostr(v):
        if np.isreal(v):
            return '{:.12g}'.format(v)
        elif isinstance(v, (string_types, np.str_)):
            return '"{}"'.format(v)

    def createsubarr(arr):
        outstr = ""
        if isinstance(arr, np.ndarray):
            if len(arr) == 1 and not isinstance(arr[0], np.ndarray):
                return '{' + tostr(arr[0]) + '},'
            outstr += '{'
            for row in arr:
                outstr += createsubarr(row)
            outstr = outstr.strip(',') + '},'
            return outstr
        else:
            return outstr + tostr(arr) + ','

    if isinstance(arr, np.ndarray) and not arr.shape:
        return tostr(arr.tolist())
    elif isinstance(arr, np.ndarray):
        return createsubarr(arr).strip(',')
    elif isinstance(arr, float):
        return tostr(arr)
    else:
        return str(arr)


class AnyPyProcessOutput(collections.OrderedDict):
    """Subclassed OrderedDict which supports partial key access."""

    def __getitem__(self, key):
        try:
            return super(AnyPyProcessOutput, self).__getitem__(key)
        except KeyError as e:
            key = _get_first_key_match(key, super(AnyPyProcessOutput, self).keys())

        try:
            return super(AnyPyProcessOutput, self).__getitem__(key)
        except KeyError as e:
            msg = 'The key {} could not be found in the data'.format(key)
            raise raise_from(KeyError(msg), None)

    def _repr_gen(self, prefix):
        items = self.items()
        if not items:
            yield prefix + '{}'
            return

        indent = prefix + '{'
        for i, (key, val) in enumerate(items):
            if i == len(self.keys()) - 1:
                end = '}'
            else:
                end = ','
            key_str = "'" + key + "'" + ': '
            val_str = _pprint.pformat(val)
            if len(prefix) + len(key_str) + len(val_str) < 80:
                yield indent + key_str + val_str + end
            else:
                yield indent + key_str
                indent = prefix + '   '
                for l in val_str.split('\n'):
                    yield indent + l if l.endswith(',') else indent + l + end
            indent = prefix + ' '

    def __repr__(self, _repr_running={}, prefix=''):
        call_key = id(self), _get_ident()
        if _repr_running is None:
            _repr_running = {}
        if call_key in _repr_running:
            return '...'
        _repr_running[call_key] = 1
        try:
            if self is None:
                return '%s()' % (self.__class__.__name__,)
            return '\n'.join(self._repr_gen(prefix))
        finally:
            del _repr_running[call_key]


TRIPEL_QUOTE_WRAP = re.compile(r'([^\[\]",\s]+)')


def _parse_data(val):
    """Convert a str AnyBody data repr into Numpy array."""
    if val.startswith('{') and val.endswith('}'):
        val = val.replace('{', '[').replace('}', ']')
    try:
        out = literal_eval(val)
    except (SyntaxError, ValueError):
        if val == '[...]':
            val = '...'
        val, _ = TRIPEL_QUOTE_WRAP.subn(r"'''\1'''", val)
        if val == '':
            val = 'None'
        if val.startswith('"') and val.endswith('"'):
            val = "'''" + val[1:-1] + r"'''"
        out = literal_eval(val)
    if isinstance(out, list):
        out = np.array(out)
    return out


ABOVE_NORMAL_PRIORITY_CLASS = 0x8000
BELOW_NORMAL_PRIORITY_CLASS = 0x4000
IDLE_PRIORITY_CLASS = 0x0040
NORMAL_PRIORITY_CLASS = 0x0020


ERROR_PATTERN = re.compile(r'^((ERROR)|(Model loading skipped)).*$', flags=re.IGNORECASE | re.M)
WARNING_PATTERN = re.compile(r'^(WARNING).*$', flags=re.IGNORECASE | re.M)
DUMP_PATTERN = re.compile(r'^(Main.*?)\s=\s(.*?(?:\n\s\s.*?)*);', flags=re.M)


def parse_anybodycon_output(raw, errors_to_ignore=None,
                            warnings_to_include=None,
                            fatal_warnings=False):
    """ Parse the output log file from AnyBodyConsole to
        for data, errors and warnings. If fatal_warnins is
        True, then warnings are also added to the error list.
    """
    warnings_to_include = warnings_to_include or []
    errors_to_ignore = errors_to_ignore or []
    output = AnyPyProcessOutput()
    # Find all data in logfile
    for dump in DUMP_PATTERN.finditer(raw):
        name, val = dump.group(1), dump.group(2)
        try:
            val = _parse_data(dump.group(2))
        except (SyntaxError, ValueError):
            warnings.warn('\n\nCould not parse console output:\n' + name)
        output[name] = val
    error_list = []

    def _add_non_ignored_errors(error_line):
        for ignored_err in errors_to_ignore:
            if ignored_err in error_line:
                break
        else:
            error_list.append(error_line)

    # Find all errors in logfile
    for match in ERROR_PATTERN.finditer(raw):
        _add_non_ignored_errors(match.group(0))
    # Find all warnings in logfile
    warning_list = []
    for match in WARNING_PATTERN.finditer(raw):
        for case in warnings_to_include:
            if case in match.group(0):
                if fatal_warnings:
                    _add_non_ignored_errors(match.group(0))
                warning_list.append(match.group(0))
                break
    if error_list:
        output['ERROR'] = error_list
    if warning_list:
        output['WARNING'] = warning_list
    return output


def get_ncpu():
    """Return the number of CPUs in the computer."""
    from multiprocessing import cpu_count
    return cpu_count()


def silentremove(filename):
    """Remove a file ignoring cases where the file does not exits."""
    try:
        os.remove(filename)
    except OSError as e:
        if e.errno != errno.ENOENT:  # errno.ENOENT : no such file or directory
            logging.debug('Error removing file: ' + filename)
            raise  # re-raise exception if a different error occured


def _run_from_ipython():
    """Return True if run from IPython."""
    try:
        __IPYTHON__
        return True
    except NameError:
        return False


def make_hash(o):
    """Make a hash from a dictionary, list, tuple or set.

    Create the hash to any sublevel that contains
    only other hashable types (including any lists, tuples, sets, and
    dictionaries).
    http://stackoverflow.com/questions/5884066/hashing-a-python-dictionary
    """
    if isinstance(o, (set, tuple, list)):

        return hash(tuple([make_hash(e) for e in o]))

    elif not isinstance(o, dict):

        return hash(o)

    new_o = copy.deepcopy(o)
    for k, v in new_o.items():
        new_o[k] = make_hash(v)

    return hash(tuple(frozenset(sorted(new_o.items()))))


_BM_CONSTANTS_AMMR1 = {
    'ON': '1',
    'OFF': '0',
    '_MUSCLES_NONE_': '0',
    '_MUSCLES_SIMPLE_': '1',
    '_MUSCLES_3E_HILL_': '2',
    '_HAND_SIMPLE_': '0',
    '_HAND_DETAILED_': '1',
    '_LEG_MODEL_OFF_': '"OFF"',
    '_LEG_MODEL_Leg_': '"Leg"',
    '_LEG_MODEL_TLEM_': '"TLEM"',
    '_MORPH_NONE_': '0',
    '_MORPH_TRUNK_TO_LEG_': '1',
    '_MORPH_LEG_TO_TRUNK_': '2',
    '_PELVIS_DISPLAY_NONE_': '0',
    '_PELVIS_DISPLAY_LEGPELVIS_ONLY_': '1',
    '_PELVIS_DISPLAY_LEGANDTRUNKPELVIS_': '2',
    '_SCALING_CUSTOM_': '-1',
    '_SCALING_STANDARD_': '0',
    '_SCALING_UNIFORM_': '1',
    '_SCALING_LENGTHMASS_': '2',
    '_SCALING_LENGTHMASSFAT_': '3',
    '_SCALING_UNIFORM_EXTMEASUREMENTS_': '4',
    '_SCALING_LENGTHMASS_EXTMEASUREMENTS_': '5',
    '_SCALING_LENGTHMASSFAT_EXTMEASUREMENTS_': '6',
    '_SCALING_LENGTHMASSFAT_MULTIDOFS_': '7',
}

_BM_CONSTANTS = {
    'ON': '1',
    'OFF': '0',
    '_MUSCLES_NONE_': '0',
    '_MUSCLES_SIMPLE_': '1',
    '_MUSCLES_3E_HILL_': '2',
    '_LEG_MODEL_OFF_': '0',
    '_LEG_MODEL_TLEM_': '1',
    '_LEG_MODEL_TLEM1_': '1',
    '_LEG_MODEL_TLEM2_': '2',
    '_LEG_MODEL_Leg_': '3',
    '_LEG_MODEL_LEG_': '3',
    '_MORPH_NONE_': '0',
    '_MORPH_TRUNK_TO_LEG_': '1',
    '_MORPH_LEG_TO_TRUNK_': '2',
    '_PELVIS_DISPLAY_NONE_': '0',
    '_PELVIS_DISPLAY_LEGPELVIS_ONLY_': '1',
    '_PELVIS_DISPLAY_LEGANDTRUNKPELVIS_': '2',
    '_DISC_STIFFNESS_NONE_': '0',
    '_DISC_STIFFNESS_LINEAR_': '1',
    '_DISC_STIFFNESS_NONLINEAR_': '2',
    '_SCALING_CUSTOM_': '-1',
    '_SCALING_STANDARD_': '0',
    '_SCALING_UNIFORM_': '1',
    '_SCALING_LENGTHMASS_': '2',
    '_SCALING_LENGTHMASSFAT_': '3',
    '_SCALING_UNIFORM_EXTMEASUREMENTS_': '4',
    '_SCALING_LENGTHMASS_EXTMEASUREMENTS_': '5',
    '_SCALING_LENGTHMASSFAT_EXTMEASUREMENTS_': '6',
    '_SCALING_LENGTHMASSFAT_MULTIDOFS_': '7',
    'CONST_MUSCLES_NONE': '0',
    'CONST_MUSCLES_SIMPLE': '1',
    'CONST_MUSCLES_3E_HILL': '2',
    'CONST_HAND_SIMPLE': '0',
    'CONST_HAND_DETAILED': '1',
    'CONST_LEG_MODEL_OFF': '0',
    'CONST_LEG_MODEL_TLEM': '1',
    'CONST_LEG_MODEL_TLEM2': '2',
    'CONST_LEG_MODEL_Leg': '3',
    'CONST_MORPH_NONE': '0',
    'CONST_MORPH_TRUNK_TO_LEG': '1',
    'CONST_MORPH_LEG_TO_TRUNK': '2',
    'CONST_PELVIS_DISPLAY_NONE': '0',
    'CONST_PELVIS_DISPLAY_LEGPELVIS_ONLY': '1',
    'CONST_PELVIS_DISPLAY_LEGANDTRUNKPELVIS': '2',
    'CONST_DISC_STIFFNESS_NONE': '0',
    'CONST_DISC_STIFFNESS_LINEAR': '1',
    'CONST_DISC_STIFFNESS_NONLINEAR ': '2',
    'CONST_SCALING_CUSTOM': '-1',
    'CONST_SCALING_STANDARD': '0',
    'CONST_SCALING_UNIFORM': '1',
    'CONST_SCALING_LENGTHMASS': '2',
    'CONST_SCALING_LENGTHMASSFAT': '3',
    'CONST_SCALING_UNIFORM_EXTMEASUREMENTS': '4',
    'CONST_SCALING_LENGTHMASS_EXTMEASUREMENTS': '5',
    'CONST_SCALING_LENGTHMASSFAT_EXTMEASUREMENTS': '6',
    'CONST_SCALING_LENGTHMASSFAT_MULTIDOFS': '7',
}


def get_bm_constants(ammr_path=None, ammr_version=2):
    """Return the BM_CONSTANT mapping.

    It will try to locate the mapping in the AMMR:
    Body/AAUHuman/Documentation/bm_constants.py

    If that fails it will use the ammr_version
    specification to select a value.
    """
    bm_constants = None
    if ammr_path is not None:
        filename = os.path.join(ammr_path, "Body/AAUHuman/Documentation/bm_constants.py")
        try:
            with open(filename) as fh:
                bm_constants = literal_eval(fh.read())
        except IOError:
            pass
    if not isinstance(bm_constants, dict):
        bm_constants = _BM_CONSTANTS if ammr_version >= 2 else _BM_CONSTANTS_AMMR1
    return bm_constants


def replace_bm_constants(d, bm_constants=None):
    """Replace BM constants with value represenation."""
    if not bm_constants:
        bm_constants = _BM_CONSTANTS

    for k, v in d.items():
        if v in bm_constants:
            d[k] = bm_constants[v]
    return d



# -*- coding: utf-8 -*-
"""
Created on Mon Sep  1 12:44:36 2014.

@author: Morten
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import * # noqa

import os
import re
import ast
import glob
import shutil
import argparse
import itertools
import contextlib
from traceback import format_list, extract_tb

import h5py
import pytest

from anypytools import AnyPyProcess, macro_commands
from anypytools.tools import (
    get_anybodycon_path, replace_bm_constants, get_bm_constants,
    anybodycon_version, find_ammr_path, get_tag, get_ammr_version,
    wraptext
)


@contextlib.contextmanager
def cwd(path):
    oldpwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(oldpwd)


class AnyTestSession(object):
    """Class for storing configuation of the AnyTest plugin to pytest."""

    def __init__(self):
        self.ammr_version = ''
        self.ams_version = ''
        self.basefolder = ''
        self.anytest_compare_dir = ''
        self.current_run_folder = None
        self.save = ''
        self.last_number = None
        self.last_session = None
        self.compare_session = None

    def configure(self, config):
        """Configure the AnyTestSession object.

        This can't be in __init__()
        since it is instantiated and added to the pytest namespace very
        early in the pytest startup.
        """
        self.basefolder = config.getoption("--anytest-storage")
        if not os.path.exists(self.basefolder):
            os.makedirs(self.basefolder)
        self.save = config.getoption(
            "--anytest-save") or config.getoption("--anytest-autosave")
        ammr_path = find_ammr_path(config.getoption("--ammr") or config.rootdir.strpath)
        self.ammr_version = get_ammr_version(ammr_path)
        self.ams_path = config.getoption("--anybodycon") or get_anybodycon_path()
        self.ams_path = os.path.abspath(self.ams_path) if self.ams_path else ''
        self.ams_version = anybodycon_version(self.ams_path)
        major_ammr_ver = 1 if self.ammr_version.startswith("1") else 2
        self.bm_constants_map = get_bm_constants(ammr_path=ammr_path, ammr_version=major_ammr_ver)
        self.compare_session = self.get_compare_session(config)
        self.run_compare_test = bool(self.save or self.compare_session)

    def get_compare_session(self, config):
        """Get the session to compare against."""
        comp = config.getoption("--anytest-compare")
        if not comp:
            return None
        elif isinstance(comp, str) and comp.isdigit():
            session = self._get_storage_folder(int(comp))
        else:
            session = self._get_storage_folder()
        if not session:
            raise ValueError(
                'Could not find any stored test runs to compare against')
        return session

    def finalize(self, config):
        """Finalize a session."""
        if self.save:
            storage_folder = os.path.join(self.basefolder, self.session_name)
            if os.path.exists(self.current_run_folder):
                shutil.copytree(self.current_run_folder, storage_folder)

    @property
    def session_name(self):
        """Return the session name if class configured to save data."""
        if self.save:
            return '{:0>4d}_{}'.format(self.last_number + 1, self.save)

    def get_compare_params(self):
        """Return (base, h5) for every file compare store."""
        if self.compare_session is None:
            return []
        with cwd(self.compare_session):
            stored_h5files = glob.glob('**/*.anydata.h5')
        return zip([self.compare_session] * len(stored_h5files), stored_h5files)

    def _get_largest_prefix(self):
        """Return the heights prefix number for folders in the self.basedir."""
        subdirs = next(os.walk(self.basefolder))[1]
        prefixes = [s.split('_')[0] for s in subdirs]
        numbers = [int(s) for s in prefixes if s.isdigit()]
        return max(numbers + [0])

    def _get_storage_folder(self, number=None):
        """Return the full folder name starting with num in self.basedir."""
        if not number:
            number = self._get_largest_prefix()
        if number is None:
            return None
        folder = glob.glob('{}\\{:0>4d}_*\\'.format(self.basefolder, number))
        if len(folder) > 1:
            raise ValueError('More folders with the same'
                             ' number prefix in {}'.format(self.basefolder))
        elif len(folder) == 0:
            return None
        else:
            return folder[0]

    def get_compare_fname(self, name, id, study):
        """Return the name of the compare h5file, and ensure the parent folder exists."""
        # Initialize and empty the current_run folder.
        if not self.current_run_folder:
            self.current_run_folder = os.path.join(
                self.basefolder, 'current_run')
            if os.path.exists(self.current_run_folder):
                shutil.rmtree(self.current_run_folder)
        if id > 0:
            compare_test_name = '{}_{}'.format(name, id)
        else:
            compare_test_name = '{}'.format(name)
        compare_test_folder = os.path.join(
            self.current_run_folder, compare_test_name)
        studyname = '{}.anydata.h5'.format(study)
        return os.path.join(compare_test_folder, studyname)


def _limited_traceback(excinfo):
    """Return a formatted traceback with this frame up removed.

    The function removes all the stack from this frame up
    (i.e from __file__ and up)
    """
    tb = extract_tb(excinfo.tb)
    try:
        idx = [__file__ in e for e in tb].index(True)
        return format_list(tb[idx + 1:])
    except ValueError:
        return format_list(tb)


def _read_header(fpath):
    """Read the commented header of anyscript test file.

    The function remvoes any leading '//' comments.
    """
    code = ''
    with open(fpath) as f:
        for line in f.readlines():
            if line.startswith('//'):
                line = line.strip('//')
                code += line
            else:
                break
    return code


@contextlib.contextmanager
def change_dir(path):
    """Context manager for changing directories."""
    prev_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)


def pytest_generate_tests(metafunc):
    if 'anytest_compare' in metafunc.fixturenames:
        params = pytest.anytest.get_compare_params()
        metafunc.parametrize("anytest_compare",
                             params, indirect=True)


@pytest.fixture
def anytest_compare(request):
    storage_folder, h5file = request.param
    current_folder = pytest.anytest.current_run_folder
    current_fname = os.path.join(current_folder, h5file)
    if os.path.exists(current_fname):
        current_h5 = h5py.File(current_fname)
        stored_h5 = h5py.File(os.path.join(storage_folder, h5file))
        yield current_h5, stored_h5
    else:
        pytest.skip('No matching h5 file found')
        yield (None, None)


def pytest_collect_file(parent, path):
    """Collect AnyScript test files."""
    if path.ext.lower() == ".any" and path.basename.lower().startswith("test_"):
        return AnyFile(path, parent)


def _format_switches(defs):
    if isinstance(defs, dict):
        defs = [defs]
    elif isinstance(defs, tuple):
        combinations = list(itertools.product(*defs))
        defs = []
        for elem in combinations:
            defs.append({k: v for d in elem for k, v in d.items()})
    elif isinstance(defs, list):
        pass
    else:
        defs = [{}]
    if len(defs) == 0:
        defs = [{}]
    return defs


def _as_absolute_paths(d, start=os.getcwd()):
    return {k: os.path.abspath(os.path.relpath(v, start)) for k, v in d.items()}


HEADER_ENSURES = (
    ('define', (dict, list, tuple)),
    ('path', (dict, list, tuple)),
    ('ignore_errors', (list, )),
    ('warnings_to_include', (list, )),
    ('fatal_warnings', (bool, )),
    ('keep_logfiles', (bool, )),
    ('logfile_prefix', (str, )),
    ('expect_errors', (list, )),
)


def _parse_header(header):
    ns = {}
    try:
        exec(header, globals(), ns)
    except SyntaxError:
        pass
    if len(ns) == 0:
        try:
            ns['define'] = ast.literal_eval(header)
        except SyntaxError:
            pass
    for name, types in HEADER_ENSURES:
        if name in ns and not isinstance(ns[name], types):
            typestr = ', '.join([t.__name__ for t in types])
            msg = '{} must be one of the following type(s) ({})'.format(
                name, typestr)
            raise TypeError(msg)
    return ns


def pytest_collection_modifyitems(session, config, items):
    """Order test so compare tests are executed last."""
    first = []
    last = []
    other = []
    for item in items:
        if hasattr(item, 'fixturenames') and 'anytest_compare' in item.fixturenames:
            last.append(item)
        elif item.get_marker('stores_h5'):
            first.append(item)
        else:
            other.append(item)
    items[:] = first + other + last


def pytest_collection_finish(session):
    """Print the AnyBodyCon executable used in the test."""
    print('\nUsing AnyBodyCon: ', pytest.anytest.ams_path)


def pytest_namespace():
    """Add an instance of the AnyTestSession class to the pytest name space."""
    return {'anytest': AnyTestSession()}


def pytest_configure(config):
    """Configure the AnyTest framework."""
    pytest.anytest.configure(config)


def pytest_unconfigure(config):
    """Finialize the test session."""
    pytest.anytest.finalize(config)


def write_macro_file(path, name, macro):
    """Write list of macros to a file."""
    filename = os.path.join(path, name + '.anymcr')
    with open(filename, 'w') as f:
        f.writelines([str(mcr) + '\n' for mcr in macro])
    return filename


class AnyFile(pytest.File):
    """pytest.File subclass for AnyScript files."""

    def collect(self):
        """Yield test cases from a AnyScript test file."""
        # Collect define statements from the header
        strheader = _read_header(self.fspath.strpath)
        header = _parse_header(strheader)
        def_list = _format_switches(header.pop('define', {}))
        def_list = [
            replace_bm_constants(d, pytest.anytest.bm_constants_map)
            for d in def_list
        ]
        path_list = _format_switches(header.pop('path', {}))
        combinations = itertools.product(def_list, path_list)
        # Run though the defines an create a test case for each
        for i, (defs, paths) in enumerate(combinations):
            if isinstance(defs, dict) and isinstance(paths, dict):
                yield AnyItem(name=self.fspath.basename, id=i, parent=self,
                              defs=defs, paths=paths, **header)
            else:
                raise ValueError('Malformed input: ', header)


class AnyItem(pytest.Item):
    """pytest.Item subclass representing individual collected tests."""

    def __init__(self, name, id, parent, defs, paths, **kwargs):
        test_name = '{}_{}'.format(name, id)
        super().__init__(test_name, parent)
        self.defs = defs
        for k, v in (self.config.getoption("define_kw") or {}):
            self.defs[k] = v
        self.defs['TEST_NAME'] = '"{}"'.format(test_name)
        if self.config.getoption("--ammr"):
            paths['AMMR_PATH'] = self.config.getoption("--ammr")
            paths['ANYBODY_PATH_AMMR'] = self.config.getoption("--ammr")
        self.paths = _as_absolute_paths(paths, start=self.config.rootdir.strpath)
        self.name = test_name
        self.expect_errors = kwargs.get('expect_errors', [])
        self.compare_study = kwargs.get('compare_study', None)
        if self.compare_study:
            self.add_marker('stores_h5')
        self.timeout = self.config.getoption("--timeout")
        self.errors = []
        self.macro = [macro_commands.Load(self.fspath.strpath, self.defs, self.paths)]
        self.compare_filename = None
        self.macro_file = None
        self.anybodycon_path = pytest.anytest.ams_path
        self.app_opts = {
            'return_task_info': True,
            'silent': True,
            'anybodycon_path': self.anybodycon_path,
            'timeout': self.timeout,
            'ignore_errors': kwargs.get('ignore_errors', []),
            'warnings_to_include': kwargs.get('warnings_to_include', []),
            'fatal_warnings': kwargs.get('fatal_warnings', False),
            'keep_logfiles': kwargs.get('keep_logfiles', False),
            'logfile_prefix': kwargs.get('logfile_prefix', None),
        }
        if not self.config.getoption("--only-load"):
            self.macro.append(macro_commands.OperationRun('Main.RunTest'))
        if pytest.anytest.run_compare_test and self.compare_study:
            # Add compare test to the test macro
            self.compare_filename = pytest.anytest.get_compare_fname(
                name, id, self.compare_study)
            save_str = 'classoperation {}.Output "Save data" --type="Deep" --file="{}"'
            save_str = save_str.format(
                self.compare_study, self.compare_filename)
            self.macro.append(macro_commands.MacroCommand(save_str))

    def runtest(self):
        """Run an AnyScript test item."""
        tmpdir = self.config._tmpdirhandler.mktemp(self.name)
        if self.compare_filename:
            os.makedirs(os.path.dirname(self.compare_filename))
        with change_dir(tmpdir.strpath):
            app = AnyPyProcess(**self.app_opts)
            result = app.start_macro(self.macro)[0]
            self.app = app
        # Ignore error due to missing Main.RunTest
        if 'ERROR' in result:
            for i, err in enumerate(result['ERROR']):
                runtest_errros = ('Error : Main.RunTest : Unresolved object',
                                  'Main.RunTest : Select Operation is not expected')
                if any(s in err for s in runtest_errros):
                    del result['ERROR'][i]
                    break
        # Check that the expected errors are present
        if self.expect_errors:
            error_list = result.get('ERROR', [])
            for xerr in self.expect_errors:
                xerr_found = False
                for i, error in enumerate(error_list):
                    if xerr in error:
                        xerr_found = True
                        del error_list[i]
                if not xerr_found:
                    self.errors.append('TEST ERROR: Expected error not '
                                       'found: "{}"'.format(xerr))
        # Add remaining errors to item's error list
        if 'ERROR' in result and len(result['ERROR']) > 0:
            self.errors.extend(result['ERROR'])
        # Add info to the hdf5 file if compare output was set
        if self.compare_filename is not None:
            import h5py
            f = h5py.File(self.compare_filename, 'a')
            f.attrs['anytest_processtime'] = result['task_processtime']
            f.attrs['anytest_macro'] = '\n'.join(result['task_macro'][:-1])
            f.attrs['anytest_ammr_version'] = pytest.anytest.ammr_version
            f.attrs['anytest_ams_version'] = pytest.anytest.ams_version
            f.close()
            basedir = os.path.dirname(self.compare_filename)
            macrofile = write_macro_file(
                basedir, self.compare_study, self.macro[:-1])
            with open(os.path.join(basedir, self.compare_study + '.bat'), 'w') as f:
                anybodygui = re.sub(r"(?i)anybodycon\.exe",
                                    r"anybody\.exe", self.anybodycon_path)
                f.write('"{}" -m "{}"'.format(anybodygui, macrofile))

        if len(self.errors) > 0:
            if self.config.getoption("--create-macros"):
                macro_name = write_macro_file(
                    self.fspath.dirname, self.name, self.macro)
                self.macro_file = macro_name
            raise AnyException(self)
        return

    def repr_failure(self, excinfo):
        """Print a representation when a test failes."""
        if isinstance(excinfo.value, AnyException):
            rtn = "Main file:\n"
            rtn += wraptext(self.fspath.strpath, initial_indent='  ')
            rtn += "\nSpecial model configuration:"
            for k, v in self.defs.items():
                rtn += "\n  #define {} {}".format(k, v)
            for k, v in self.paths.items():
                rtn += "\n  #path {} {}".format(k, v)
            if self.macro_file is not None:
                macro_file = self.macro_file.replace(os.sep, os.altsep)
                rtn += '\nMacro:'
                rtn += wraptext('\n  anybody.exe -m "{}" &'.format(macro_file), initial_indent='  ')
            rtn += '\nErrors:'
            for elem in self.errors:
                rtn += '\n'
                rtn += wraptext(elem, initial_indent='> ', subsequent_indent='  ')
            return rtn
        else:
            return str(excinfo.value)

    def reportinfo(self):
        return self.fspath, 0, "AnyBody Simulation: %s" % self.name


class AnyException(Exception):
    """Custom exception for error reporting."""


def parse_save_name(stringval):
    """Parse function."""
    if not stringval:
        raise argparse.ArgumentTypeError("Argument can't be empty.")
    not_allowed = ''.join(c for c in r"\/:*?<>|" if c in stringval)
    if not_allowed:
        raise argparse.ArgumentTypeError("Characters are not allowed: "
                                         "/:*?<>|\\ (it has %r)" % not_allowed)
    return stringval


def pytest_addoption(parser):
    group = parser.getgroup("anypytools", "testing AnyBody models")

    group.addoption("--anybodycon", action="store", metavar="path",
                    help="anybodycon.exe used in test: default or "
                    "path-to-anybodycon")
    group.addoption("--ammr", action="store", metavar="path",
                    help="Can be used to specify which AnyBody Managed Model "
                    "Repository (AMMR) to use. Setting this will pass a "
                    "'AMMR_PATH' path statement for all models")
    group.addoption("--only-load", action="store_true",
                    help="Only run a load test. I.e. do not run the "
                    "'RunTest' macro")
    group.addoption("--define", action='append',
                    type=lambda kv: kv.split("=", 1), dest='define_kw',
                    help="Add custom define statements parse to all AnyScript models. "
                    "Must be given in the form: --define MYDEF=6")
    group._addoption("--timeout", default=3600, type=int,
                     help="terminate tests after a certain timeout period")
    tag = get_tag()
    group.addoption("--anytest-save", metavar="NAME", type=parse_save_name,
                    help='Save the current run into folder '
                    '`~/.anytest/counter-NAME/`. Default: `<commitid>_'
                    '<date>_<time>_<isdirty>`, example: `%s`.' % tag)
    group.addoption("--anytest-autosave", action='store_const', const=tag,
                    help="Autosave the current run into folder"
                    "'~/.anytest/counter_%s/" % tag)
    group.addoption("--create-macros", action="store_true",
                    help="Create a macro file if the test fails. This makes it "
                    "easy to re-run the failed test in the gui application.")
    group.addoption("--anytest-compare",
                    metavar="NUM", nargs="?", default=[], const=True,
                    help="Compare the current run against run NUM or the latests "
                    "saved run if unspecified.")
    group.addoption("--anytest-storage",
                    metavar="path", default=os.path.expanduser("~/.anytest"),
                    help="Specify a path to store the runs (when --anytest-save "
                    "or --benchmark-autosave are used). Default: %(default)r."
                    )



from anypytools.abcutils import AnyPyProcess

nSteplist =  [10,20,40,50,60,70,80,90,100,110,120]
app = AnyPyProcess(num_processes = 2)

macrolist = []
for nStep in nSteplist:
    macrolist.append( ['load "Demo.Arm2D.any" -def N_STEP="{0}"'.format(nStep),
                       'operation Main.Study.InverseDynamics', 'run'] )
    
app.start_macro(macrolist)




# -*- coding: utf-8 -*-
"""
Created on Sun Sep  7 13:25:38 2014.

@author: Morten
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import (ascii, bytes, chr, dict, filter, hex, input,  # noqa
                      int, map, next, oct, open, pow, range, round,
                      str, super, zip)

import logging
import collections
from copy import deepcopy

import numpy as np
from datashape import discover
from datashape.coretypes import (Record, var, Option)
try:
    from odo import convert
except ImportError:
    from into import convert

from dynd import nd

from anypytools.tools import AnyPyProcessOutputList

logger = logging.getLogger('abt.anypytools')


@convert.register(nd.array, AnyPyProcessOutputList, cost=1.0)
def convert(res, **kwargs):
    # Hack to ensure values are always interpreted as floats
    # This is necessary because the dump command in AnyBody
    # may cause whole number float to look like int.
    res = deepcopy(res)
    for elem in res:
        for key in elem:
            if key.startswith('Main') and isinstance(elem[key], np.ndarray):
                try:
                    elem[key] = elem[key].astype('float')
                except ValueError:
                    pass
    ###
    prepared_data, ds = convert_and_extract_dshape(res, **kwargs)
    return nd.array(prepared_data, dtype=str(ds))


def convert_to_nested_structure(obj):
    assert isinstance(obj, collections.Mapping)
    for key in obj:
        if '.' in key:
            first, last = key.split('.', 1)
            if first not in obj:
                obj[first] = collections.OrderedDict()
            obj[first][last] = obj.pop(key)
            convert_to_nested_structure(obj[first])


def remove_dots_in_key_names(obj):
    assert isinstance(obj, collections.Mapping)
    for key in list(obj.keys()):
        obj[key.replace('.', '_')] = obj.pop(key)


def convert_items_to_python_types(obj):
    assert isinstance(obj, collections.Mapping)
    for k in obj:
        if isinstance(obj[k], collections.Mapping):
            convert_items_to_python_types(obj[k])
        else:
            if isinstance(obj[k], np.ndarray):
                obj[k] = obj.pop(k).tolist()
            if not obj[k] and isinstance(obj[k], collections.Iterable):
                obj[k] = None


def update_dict_if(d, u, condition=True):
    """Update one nested dictionary 'd' with the keys from another 'u'.

    If the keys already exist, then values are only if condition
    function returns true
    """
    if condition is True:
        condition = lambda old_val, new_val: True  # noqa
    elif condition is False:
        condition = lambda old_val, new_val: False # noqa
    for k in u:
        if isinstance(u[k], collections.Mapping):
            r = update_dict_if(d.get(k, {}), u[k], condition)
            d[k] = r
        elif k in d:
            if condition(d[k], u[k]):
                d[k] = u[k]
        else:
            d[k] = u[k]
    return d


def build_datashape(sample_data):
    """Build a datashape from sample_data.

    In case of multiple dimensions, the firt is replace by 'var' to
    handle varing time samples in models
    """
    if isinstance(sample_data, collections.Mapping):
        return Record((k, build_datashape(v)) for k, v in sample_data.items())
    else:
        ds = discover(sample_data)
        if len(ds) > 1:
            ds = var * ds.subarray(1)
        else:
            ds = Option(ds)
        return ds


def clean_sample(mapping, key):
    """Scrubs the values of a nested data structure.

    The function leaves the prototype
    empty values. List are replaced by [] and values by None
    """
    if isinstance(mapping[key], collections.Mapping):
        for subkey in mapping[key]:
            clean_sample(mapping[key], subkey)
    elif isinstance(mapping[key], collections.Iterable):
        mapping[key] = []
    else:
        mapping[key] = None


def convert_and_extract_dshape(result_list, create_nested_structure=False, **kwargs):

    result_list = deepcopy(result_list)

    if create_nested_structure:
        for elem in result_list:
            convert_to_nested_structure(elem)
    else:
        for elem in result_list:
            # pass
            remove_dots_in_key_names(elem)

    for elem in result_list:
        convert_items_to_python_types(elem)

    def update_check(old_value, new_value):
        if isinstance(old_value, collections.Iterable):
            if not any(old_value):
                return True
        elif not old_value:
            return True
        else:
            return False

    # Create a sample data structure with the keys from all the results.
    sample_data = collections.OrderedDict()
    for data_structure in result_list:
        update_dict_if(sample_data, data_structure, update_check)

    # If some task info is present set this data so the discover function
    # will always find the correct data type.
    if 'task_id' in sample_data:
        sample_data['task_macro_hash'] = 42
        sample_data['task_id'] = 42
        sample_data['task_work_dir'] = "string"
        sample_data['task_name'] = "string"
        sample_data['task_processtime'] = 10.1
        sample_data['task_macro'] = ['string', 'string']
        sample_data['task_logfile'] = 'string'

    dshape = build_datashape(sample_data)

    empty_sample_data = deepcopy(sample_data)
    for key in sample_data:
        clean_sample(empty_sample_data, key)

    for elem in result_list:
        # Update_dict_if( ) with a always False condition, will
        # only add missing keys
        update_dict_if(elem, empty_sample_data, condition=False)

    return (result_list, dshape)



# -*- coding: utf-8 -*-
"""AnyPyTools library."""

import sys
import platform
import logging

from anypytools.abcutils import AnyPyProcess, execute_anybodycon
from anypytools.macroutils import AnyMacro
from anypytools import macro_commands
from anypytools.tools import (ABOVE_NORMAL_PRIORITY_CLASS, BELOW_NORMAL_PRIORITY_CLASS,
                              IDLE_PRIORITY_CLASS, NORMAL_PRIORITY_CLASS)


logger = logging.getLogger('abt.anypytools')
logger.addHandler(logging.NullHandler())


__all__ = [
    'datautils', 'h5py_wrapper', 'AnyPyProcess', 'AnyMacro', 'macro_commands',
    'print_versions', 'execute_anybodycon', 'ABOVE_NORMAL_PRIORITY_CLASS',
    'BELOW_NORMAL_PRIORITY_CLASS', 'IDLE_PRIORITY_CLASS', 'NORMAL_PRIORITY_CLASS'
]

__version__ = '0.13.0'


def print_versions():
    """Print all the versions of software that AnyPyTools relies on."""
    import numpy as np
    import scipy as sp
    print("-=" * 38)
    print("AnyPyTools version: %s" % __version__)
    print("NumPy version: %s" % np.__version__)
    print("SciPy version: %s" % sp.__version__)
    print("Python version: %s" % sys.version)
    (sysname, nodename, release, version, machine, processor) = \
        platform.uname()
    print("Platform: %s-%s-%s (%s)" % (sysname, release, machine, version))
    if sysname == "Linux":
        print("Linux dist: %s" % " ".join(platform.linux_distribution()[:-1]))
    if not processor:
        processor = "not recognized"
    print("Processor: %s" % processor)
    print("Byte-ordering: %s" % sys.byteorder)
    print("-=" * 38)
